import argparse
import random

import terminaltexteffects.utils.argtypes as argtypes
from terminaltexteffects.base_character import EffectCharacter, EventHandler
from terminaltexteffects.utils.terminal import Terminal
from terminaltexteffects.utils import graphics, motion, argtypes, easing


def add_arguments(subparsers: argparse._SubParsersAction) -> None:
    """Adds arguments to the subparser.

    Args:
        subparser (argparse._SubParsersAction): subparser to add arguments to
    """
    effect_parser = subparsers.add_parser(
        "swarm",
        formatter_class=argtypes.CustomFormatter,
        help="Characters are grouped into swarms and move around the terminal before settling into position.",
        description="Characters are grouped into swarms and move around the terminal before settling into position.",
        epilog=f"""Example: terminaltexteffects swarm -a 0.01 --base-color 0f74fa --flash-color f2ea79 --final-color 96ed89 --swarm-size 5 --swarm-coordination 80""",
    )
    effect_parser.set_defaults(effect_class=SwarmEffect)
    effect_parser.add_argument(
        "-a",
        "--animation-rate",
        type=argtypes.valid_animationrate,
        default=0.01,
        help="Time, in seconds, between animation steps.",
    )
    effect_parser.add_argument(
        "--base-color",
        type=argtypes.valid_color,
        default="0f74fa",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the characters when not flashing.",
    )
    effect_parser.add_argument(
        "--flash-color",
        type=argtypes.valid_color,
        default="f2ea79",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the character flash. Characters flash when moving.",
    )
    effect_parser.add_argument(
        "--final-color",
        type=argtypes.valid_color,
        default="96ed89",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the characters when they reach their input coordinate.",
    )
    effect_parser.add_argument(
        "--swarm-size",
        type=argtypes.valid_percent_greater_than_zero,
        default=5,
        help="Percent of total characters in each swarm. Must be between 1 and 100.",
    )
    effect_parser.add_argument(
        "--swarm-coordination",
        type=argtypes.valid_percent_greater_than_zero,
        default=80,
        help="Percent of characters in a swarm that move as a group. Must be between 1 and 100.",
    )


class SwarmEffect:
    """Characters behave with swarm characteristics before flying into position."""

    def __init__(self, terminal: Terminal, args: argparse.Namespace):
        self.terminal = terminal
        self.args = args
        self.pending_chars: list[EffectCharacter] = []
        self.animating_chars: list[EffectCharacter] = []
        self.swarms: list[list[EffectCharacter]] = []
        self.swarm_size: int = max(round(len(self.terminal.characters) * (self.args.swarm_size / 100)), 1)

    def make_swarms(self) -> None:
        unswarmed_characters = list(self.terminal.characters[::-1])
        while unswarmed_characters:
            new_swarm: list[EffectCharacter] = []
            for _ in range(self.swarm_size):
                if unswarmed_characters:
                    new_swarm.append(unswarmed_characters.pop())
                else:
                    break
            self.swarms.append(new_swarm)

    def prepare_data(self) -> None:
        """Prepares the data for the effect by creating swarms of characters and setting waypoints and animations."""
        self.make_swarms()
        final_gradient = graphics.Gradient(self.args.base_color, self.args.final_color, 10)
        swarm_gradient = graphics.Gradient(self.args.base_color, self.args.flash_color, 7)
        flash_list = [self.args.flash_color for _ in range(10)]
        swarm_gradient_mirror = list(swarm_gradient) + flash_list + list(swarm_gradient)[::-1]
        for swarm in self.swarms:
            swarm_area_coordinate_map: dict[motion.Coord, list[motion.Coord]] = {}
            swarm_spawn = self.terminal.random_coord(outside_scope=True)
            swarm_areas: list[motion.Coord] = []
            swarm_area_count = random.randint(2, 4)
            # create areas where characters will swarm
            while len(swarm_areas) < swarm_area_count:
                # get random coord within inner 90% of terminal
                col = random.randint(round(self.terminal.input_width * 0.1), round(self.terminal.input_width * 0.9) + 1)
                row = random.randint(
                    round(self.terminal.output_area.top * 0.1), round(self.terminal.output_area.top * 0.9) + 1
                )
                area = motion.Coord(col, row)
                if area not in swarm_areas:
                    swarm_areas.append(area)
                    # swarm area radius is a function of the terminal size, approximately 1/4 of the smallest dimension
                    swarm_area_coordinate_map[area] = motion.Motion.find_coords_in_circle(
                        area,
                        max(min(self.terminal.output_area.right, self.terminal.output_area.top) // 4, 1),
                        self.swarm_size,
                    )

            # assign characters waypoints for swarm areas and inner waypoints within the swarm areas
            for character in swarm:
                swarm_area_count = 0
                character.motion.set_coordinate(swarm_spawn)
                flash_scn = character.animation.new_scene("speed", sync=graphics.SyncMetric.DISTANCE)
                for step in swarm_gradient_mirror:
                    flash_scn.add_frame(character.input_symbol, 1, color=step)
                for _, swarm_area_coords in swarm_area_coordinate_map.items():
                    swarm_area_name = f"{swarm_area_count}_swarm_area"
                    swarm_area_count += 1
                    origin_wpt = character.motion.new_waypoint(
                        swarm_area_name,
                        random.choice(swarm_area_coords),
                        speed=0.25,
                        ease=easing.out_sine,
                    )
                    character.event_handler.register_event(
                        EventHandler.Event.WAYPOINT_ACTIVATED, origin_wpt, EventHandler.Action.ACTIVATE_SCENE, flash_scn
                    )
                    character.event_handler.register_event(
                        EventHandler.Event.WAYPOINT_ACTIVATED, origin_wpt, EventHandler.Action.SET_LAYER, 1
                    )
                    character.event_handler.register_event(
                        EventHandler.Event.WAYPOINT_REACHED, origin_wpt, EventHandler.Action.DEACTIVATE_SCENE, flash_scn
                    )
                    inner_waypoints = 0
                    total_inner_waypoints = 2
                    while inner_waypoints < total_inner_waypoints:
                        next_coord = random.choice(swarm_area_coords)
                        inner_waypoints += 1
                        character.motion.new_waypoint(
                            str(len(character.motion.waypoints)),
                            next_coord,
                            speed=0.1,
                            ease=easing.in_out_sine,
                        )
                # create landing waypoint and scene
                input_wpt = character.motion.new_waypoint(
                    "input_coord", character.input_coord, speed=0.3, ease=easing.in_out_quad
                )
                input_scn = character.animation.new_scene("input")
                for step in final_gradient:
                    input_scn.add_frame(character.input_symbol, 3, color=step)
                character.event_handler.register_event(
                    EventHandler.Event.WAYPOINT_REACHED, input_wpt, EventHandler.Action.ACTIVATE_SCENE, input_scn
                )
                character.event_handler.register_event(
                    EventHandler.Event.WAYPOINT_REACHED, input_wpt, EventHandler.Action.SET_LAYER, 0
                )
                # create flash effect when moving between waypoints
                character.event_handler.register_event(
                    EventHandler.Event.WAYPOINT_ACTIVATED, input_wpt, EventHandler.Action.ACTIVATE_SCENE, flash_scn
                )
                # chain waypoints
                last_wpt = None
                for wpt in character.motion.waypoints.values():
                    if last_wpt:
                        character.event_handler.register_event(
                            EventHandler.Event.WAYPOINT_REACHED, last_wpt, EventHandler.Action.ACTIVATE_WAYPOINT, wpt
                        )
                    last_wpt = wpt

    def run(self) -> None:
        """Runs the effect."""
        self.prepare_data()
        call_next = True
        active_swarm_area = "0_swarm_area"
        while self.swarms or self.animating_chars:
            if self.swarms and call_next:
                call_next = False
                current_swarm = self.swarms.pop()
                active_swarm_area = "0_swarm_area"
                for character in current_swarm:
                    character.motion.activate_waypoint(character.motion.waypoints["0_swarm_area"])
                    character.is_active = True
                    self.animating_chars.append(character)
            self.terminal.print()
            self.animate_chars()
            if len(self.animating_chars) < len(current_swarm):
                call_next = True
            if current_swarm:
                for character in current_swarm:
                    if (
                        character.motion.active_waypoint
                        and character.motion.active_waypoint.waypoint_id != active_swarm_area
                        and "swarm_area" in character.motion.active_waypoint.waypoint_id
                        and int(character.motion.active_waypoint.waypoint_id[0]) > int(active_swarm_area[0])
                    ):
                        active_swarm_area = character.motion.active_waypoint.waypoint_id
                        for other in current_swarm:
                            if other is not character and random.random() < (self.args.swarm_coordination / 100):
                                other.motion.activate_waypoint(other.motion.waypoints[active_swarm_area])
                        break

            # remove completed chars from animating chars
            self.animating_chars = [
                animating_char
                for animating_char in self.animating_chars
                if not animating_char.animation.active_scene_is_complete()
                or not animating_char.motion.movement_is_complete()
            ]

    def animate_chars(self) -> None:
        """Animates the characters by calling the move method and step animation. Move characters prior to stepping animation
        to ensure waypoint synced animations have the latest waypoint progress information."""
        for animating_char in self.animating_chars:
            animating_char.motion.move()
            animating_char.animation.step_animation()



import argparse
import random
import terminaltexteffects.utils.argtypes as argtypes
from terminaltexteffects.base_character import EffectCharacter, EventHandler
from terminaltexteffects.utils.terminal import Terminal
from terminaltexteffects.utils import graphics, motion, argtypes, easing


def add_arguments(subparsers: argparse._SubParsersAction) -> None:
    """Adds arguments to the subparser.

    Args:
        subparser (argparse._SubParsersAction): subparser to add arguments to
    """
    effect_parser = subparsers.add_parser(
        "blackhole",
        formatter_class=argtypes.CustomFormatter,
        help="Characters are consumed by a black hole and explode outwards.",
        description="Characters are consumed by a black hole and explode outwards.",
        epilog=f"""
Example: terminaltexteffects blackhole -a 0.01 --star-colors ffcc0d ff7326 ff194d bf2669 702a8c 049dbf --final-color 00a7c2""",
    )
    effect_parser.set_defaults(effect_class=BlackholeEffect)
    effect_parser.add_argument(
        "-a",
        "--animation-rate",
        type=argtypes.valid_animationrate,
        default=0.01,
        help="Time, in seconds, between animation steps.",
    )
    effect_parser.add_argument(
        "--blackhole-color",
        type=argtypes.valid_color,
        default="ffffff",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the stars that comprise the blackhole border.",
    )
    effect_parser.add_argument(
        "--star-colors",
        type=argtypes.valid_color,
        nargs="*",
        default=["ffcc0d", "ff7326", "ff194d", "bf2669", "702a8c", "049dbf"],
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="List of colors from which character colors will be chosen and applied after the explosion, but before the cooldown to final color.",
    )
    effect_parser.add_argument(
        "--final-color",
        type=argtypes.valid_color,
        default="00a7c2",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Final color that characters will shift to after exploding.",
    )


class BlackholeEffect:
    """Effect that creates a blackhole in a starfield, consumes the stars, and explodes the input characters out to position."""

    def __init__(self, terminal: Terminal, args: argparse.Namespace):
        self.terminal = terminal
        self.args = args
        self.pending_chars: list[EffectCharacter] = []
        self.animating_chars: list[EffectCharacter] = []
        self.blackhole_chars: list[EffectCharacter] = []
        self.awaiting_consumption_chars: list[EffectCharacter] = []
        self.blackhole_radius = max(
            min(round(self.terminal.input_width * 0.4), round(self.terminal.input_height * 0.4)), 4
        )

    def prepare_blackhole(self) -> None:
        star_symbols = [
            "*",
            "✸",
            "✺",
            "✹",
            "✷",
            "✵",
            "✴",
            "✶",
            "⋆",
            "'",
            ".",
            "⬫",
            "⬪",
            "⬩",
            "⬨",
            "⬧",
            "⬦",
            "⬥",
        ]
        starfield_colors = graphics.Gradient("4a4a4d", "ffffff", 6).colors
        gradient_map = {}
        for color in starfield_colors:
            gradient_map[color] = graphics.Gradient(color, "000000", 10)
        available_chars = list(self.terminal.characters)
        while len(self.blackhole_chars) < self.blackhole_radius * 3 and available_chars:
            self.blackhole_chars.append(available_chars.pop(random.randrange(0, len(available_chars))))
        black_hole_ring_positions = motion.Motion.find_coords_on_circle(
            self.terminal.output_area.center, self.blackhole_radius, len(self.blackhole_chars)
        )
        for character in self.blackhole_chars:
            next_pos = black_hole_ring_positions.pop(0)
            blackhole_wpt = character.motion.new_waypoint("blackhole", next_pos, speed=0.5, ease=easing.in_out_sine)
            blackhole_scn = character.animation.new_scene("blackhole")
            blackhole_scn.add_frame("✸", 1, color=self.args.blackhole_color)
            character.event_handler.register_event(
                EventHandler.Event.WAYPOINT_ACTIVATED, blackhole_wpt, EventHandler.Action.SET_LAYER, 1
            )
        for character in self.terminal.characters:
            character.is_active = True
            starting_scn = character.animation.new_scene("starting")
            star_symbol = random.choice(star_symbols)
            star_color = random.choice(starfield_colors)
            starting_scn.add_frame(star_symbol, 1, color=star_color)
            character.animation.activate_scene(starting_scn)
            if character not in self.blackhole_chars:
                character.motion.set_coordinate(self.terminal.random_coord())
                singluarity_wpt = character.motion.new_waypoint(
                    "singularity",
                    self.terminal.output_area.center,
                    speed=0.3,
                    ease=easing.in_expo,
                )
                consumed_scn = character.animation.new_scene("consumed")
                for color in gradient_map[star_color]:
                    consumed_scn.add_frame(star_symbol, 1, color=color)
                consumed_scn.add_frame(" ", 1)
                consumed_scn.sync = graphics.SyncMetric.DISTANCE
                character.event_handler.register_event(
                    EventHandler.Event.WAYPOINT_ACTIVATED, singluarity_wpt, EventHandler.Action.SET_LAYER, 2
                )
                character.event_handler.register_event(
                    EventHandler.Event.WAYPOINT_ACTIVATED,
                    singluarity_wpt,
                    EventHandler.Action.ACTIVATE_SCENE,
                    consumed_scn,
                )
                self.awaiting_consumption_chars.append(character)
        random.shuffle(self.awaiting_consumption_chars)

    def rotate_blackhole(self) -> None:
        self.blackhole_chars = self.blackhole_chars[2:] + self.blackhole_chars[:2]
        black_hole_ring_positions = motion.Motion.find_coords_on_circle(
            self.terminal.output_area.center, self.blackhole_radius, len(self.blackhole_chars)
        )
        for character in self.blackhole_chars:
            next_pos = black_hole_ring_positions.pop(0)
            rotate_wpt = character.motion.new_waypoint("rotate", next_pos, speed=0.1)
            character.motion.activate_waypoint(rotate_wpt)
            if character not in self.animating_chars:
                self.animating_chars.append(character)

    def collapse_blackhole(self) -> None:
        black_hole_ring_positions = motion.Motion.find_coords_on_circle(
            self.terminal.output_area.center, self.blackhole_radius + 4, len(self.blackhole_chars)
        )
        unstable_symbols = ["◦", "◎", "◉", "●", "◉", "◎", "◦"]
        point_char_made = False
        for character in self.blackhole_chars:
            next_pos = black_hole_ring_positions.pop(0)
            expand_wpt = character.motion.new_waypoint("expand", next_pos, speed=0.1, ease=easing.in_expo)
            collapse_wpt = character.motion.new_waypoint(
                "collapse",
                self.terminal.output_area.center,
                speed=0.3,
                ease=easing.in_expo,
            )
            character.event_handler.register_event(
                EventHandler.Event.WAYPOINT_REACHED, expand_wpt, EventHandler.Action.ACTIVATE_WAYPOINT, collapse_wpt
            )
            if not point_char_made:
                point_scn = character.animation.new_scene("point")
                for _ in range(3):
                    for symbol in unstable_symbols:
                        point_scn.add_frame(symbol, 6, color=random.choice(self.args.star_colors))
                character.event_handler.register_event(
                    EventHandler.Event.WAYPOINT_REACHED, collapse_wpt, EventHandler.Action.ACTIVATE_SCENE, point_scn
                )
                character.event_handler.register_event(
                    EventHandler.Event.WAYPOINT_REACHED, collapse_wpt, EventHandler.Action.SET_LAYER, 3
                )
                point_char_made = True

            character.motion.activate_waypoint(expand_wpt)
            self.animating_chars.append(character)

    def explode_singularity(self) -> None:
        star_colors = ["ffcc0d", "ff7326", "ff194d", "bf2669", "702a8c" "049dbf"]
        for character in self.terminal.characters:
            nearby_coord = motion.Motion.find_coords_on_circle(character.input_coord, 4, 5)[random.randrange(0, 5)]
            nearby_wpt = character.motion.new_waypoint(
                "nearby_wpt",
                nearby_coord,
                speed=random.randint(2, 3) / 10,
                ease=easing.out_expo,
            )
            input_wpt = character.motion.new_waypoint(
                "input_wpt",
                character.input_coord,
                speed=random.randint(3, 5) / 100,
                ease=easing.in_cubic,
            )
            explode_scn = character.animation.new_scene("explode")
            explode_star_color = random.choice(star_colors)
            explode_scn.add_frame(character.input_symbol, 1, color=explode_star_color)
            cooling_scn = character.animation.new_scene("cooling")
            cooling_gradient = graphics.Gradient(explode_star_color, self.args.final_color, 10)
            for color in cooling_gradient:
                cooling_scn.add_frame(character.input_symbol, 20, color=color)
            character.event_handler.register_event(
                EventHandler.Event.WAYPOINT_REACHED, nearby_wpt, EventHandler.Action.ACTIVATE_WAYPOINT, input_wpt
            )
            character.event_handler.register_event(
                EventHandler.Event.WAYPOINT_REACHED, nearby_wpt, EventHandler.Action.ACTIVATE_SCENE, cooling_scn
            )
            character.animation.activate_scene(explode_scn)
            character.motion.activate_waypoint(nearby_wpt)
            self.animating_chars.append(character)

    def prepare_data(self) -> None:
        """Prepares the data for the effect by ___."""
        self.prepare_blackhole()

    def run(self) -> None:
        """Runs the effect."""
        self.prepare_data()
        formation_delay = max(100 // len(self.blackhole_chars), 10)
        f_delay = formation_delay
        next_char_consuming_delay = 0
        max_consume = max(min(int(0.1 * len(self.terminal.characters)), 15), 2)
        phase = "forming"
        awaiting_blackhole_chars = list(self.blackhole_chars)
        while self.animating_chars or phase != "complete":
            if phase == "forming":
                if awaiting_blackhole_chars:
                    if not f_delay:
                        next_char = awaiting_blackhole_chars.pop(0)
                        next_char.motion.activate_waypoint(next_char.motion.waypoints["blackhole"])
                        next_char.animation.activate_scene(next_char.animation.scenes["blackhole"])
                        self.animating_chars.append(next_char)
                        f_delay = formation_delay
                    else:
                        f_delay -= 1
                else:
                    if not self.animating_chars:
                        phase = "consuming"
            elif phase == "consuming":
                if self.awaiting_consumption_chars:
                    if not next_char_consuming_delay:
                        for _ in range(random.randrange(1, max_consume)):
                            if self.awaiting_consumption_chars:
                                next_char = self.awaiting_consumption_chars.pop(0)
                                next_char.motion.activate_waypoint(next_char.motion.waypoints["singularity"])
                                self.animating_chars.append(next_char)
                            else:
                                break
                        max_consume += 1
                        next_char_consuming_delay = random.randrange(0, 10)
                    else:
                        next_char_consuming_delay -= 1

                else:
                    if not self.animating_chars:
                        phase = "collapsing"
                if all([character.motion.active_waypoint is None for character in self.blackhole_chars]):
                    self.rotate_blackhole()
            elif phase == "collapsing":
                if all([character.motion.active_waypoint is None for character in self.blackhole_chars]):
                    self.collapse_blackhole()
                    phase = "exploding"
            elif phase == "exploding":
                if all(
                    [
                        character.motion.active_waypoint is None and character.animation.active_scene is None
                        for character in self.blackhole_chars
                    ]
                ):
                    self.explode_singularity()
                    phase = "complete"

            self.animate_chars()
            self.terminal.print()

            # remove completed chars from animating chars
            self.animating_chars = [
                animating_char
                for animating_char in self.animating_chars
                if not animating_char.animation.active_scene_is_complete()
                or not animating_char.motion.movement_is_complete()
            ]

    def animate_chars(self) -> None:
        """Animates the characters by calling the move method and step animation. Move characters prior to stepping animation
        to ensure waypoint synced animations have the latest waypoint progress information."""
        for animating_char in self.animating_chars:
            animating_char.motion.move()
            animating_char.animation.step_animation()



import argparse

import terminaltexteffects.utils.argtypes as argtypes
from terminaltexteffects.base_character import EffectCharacter, EventHandler
from terminaltexteffects.utils.terminal import Terminal
from terminaltexteffects.utils import graphics, argtypes


def add_arguments(subparsers: argparse._SubParsersAction) -> None:
    """Adds arguments to the subparser.

    Args:
        subparser (argparse._SubParsersAction): subparser to add arguments to
    """
    effect_parser = subparsers.add_parser(
        "waves",
        formatter_class=argtypes.CustomFormatter,
        help="Waves travel across the terminal leaving behind the characters.",
        description="Waves travel across the terminal leaving behind the characters.",
        epilog=f"""{argtypes.EASING_EPILOG}

Example: terminaltexteffects waves -a 0.01 --wave-deep-color 4651e3 --wave-shallow-color 57FB92 --final-color 80F7E6 --wave-count 6 --wave-easing IN_OUT_SINE""",
    )
    effect_parser.set_defaults(effect_class=WavesEffect)
    effect_parser.add_argument(
        "-a",
        "--animation-rate",
        type=argtypes.valid_animationrate,
        default=0.01,
        help="Time, in seconds, between animation steps.",
    )
    effect_parser.add_argument(
        "--wave-deep-color",
        type=argtypes.valid_color,
        default="4651e3",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the deepest part of the wave. A gradient is generated between the shallow and deep parts of the wave.",
    )
    effect_parser.add_argument(
        "--wave-shallow-color",  # make more descriptive
        type=argtypes.valid_color,
        default="57FB92",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the shallow part of the wave. A gradient is generated between the shallow and deep parts of the wave..",
    )
    effect_parser.add_argument(
        "--final-color",
        type=argtypes.valid_color,
        default="80F7E6",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the final character. A gradient is generated between the shallow part of the wave and the final color.",
    )
    effect_parser.add_argument(
        "--wave-count",
        type=argtypes.positive_int,
        default=6,
        help="Number of waves to generate. n > 0.",
    )
    effect_parser.add_argument(
        "--wave-easing",
        default="IN_OUT_SINE",
        type=argtypes.valid_ease,
        help="Easing function to use for wave travel.",
    )


class WavesEffect:
    """Effect that creates waves that travel across the terminal, leaving behind the characters."""

    def __init__(self, terminal: Terminal, args: argparse.Namespace):
        self.terminal = terminal
        self.args = args
        self.pending_columns: list[list[EffectCharacter]] = []
        self.animating_chars: list[EffectCharacter] = []

    def prepare_data(self) -> None:
        """Prepares the data for the effect by creating the wave animations."""
        block_wipe_start = ("▁", "▂", "▃", "▄", "▅", "▆", "▇", "█")
        block_wipe_end = ("▇", "▆", "▅", "▄", "▃", "▂", "▁")
        wave_gradient_light = graphics.Gradient(self.args.wave_shallow_color, self.args.wave_deep_color, 8)
        wave_gradient_dark = graphics.Gradient(self.args.wave_deep_color, self.args.wave_shallow_color, 8)
        wave_gradient = list(wave_gradient_light) + list(wave_gradient_dark)[1:]
        colored_waves = list(zip(block_wipe_start + block_wipe_end, wave_gradient))
        final_gradient = graphics.Gradient(self.args.wave_shallow_color, self.args.final_color, 8)
        for character in self.terminal.characters:
            wave_scn = character.animation.new_scene("wave")
            wave_scn.ease = self.args.wave_easing
            for _ in range(self.args.wave_count):
                for block, color in colored_waves:
                    wave_scn.add_frame(block, 3, color=color)
            final_scn = character.animation.new_scene("final")
            for step in final_gradient:
                final_scn.add_frame(character.input_symbol, 15, color=step)
            character.event_handler.register_event(
                EventHandler.Event.SCENE_COMPLETE, wave_scn, EventHandler.Action.ACTIVATE_SCENE, final_scn
            )
            character.animation.activate_scene(wave_scn)
        columns = self.terminal.input_by_column()
        for _, column in columns.items():
            self.pending_columns.append(column)

    def run(self) -> None:
        """Runs the effect."""
        self.prepare_data()
        while self.pending_columns or self.animating_chars:
            if self.pending_columns:
                next_column = self.pending_columns.pop(0)
                for character in next_column:
                    character.is_active = True
                    self.animating_chars.append(character)
            self.terminal.print()
            self.animate_chars()

            self.animating_chars = [
                animating_char
                for animating_char in self.animating_chars
                if not animating_char.animation.active_scene_is_complete()
                or not animating_char.motion.movement_is_complete()
            ]

    def animate_chars(self) -> None:
        """Animates the characters by calling the move method and step animation. Move characters prior to stepping animation
        to ensure waypoint synced animations have the latest waypoint progress information."""
        for animating_char in self.animating_chars:
            animating_char.motion.move()
            animating_char.animation.step_animation()



import random
import typing, math
from dataclasses import dataclass
import terminaltexteffects.utils.easing as easing

if typing.TYPE_CHECKING:
    from terminaltexteffects import base_character


@dataclass(eq=True, frozen=True)
class Coord:
    """A coordinate with row and column values.

    Args:
        column (int): column value
        row (int): row value"""

    column: int
    row: int


@dataclass
class Waypoint:
    """A coordinate, speed, and easing function.

    Args:
        waypoint_id (str): unique identifier for the waypoint
        coord (Coord): coordinate
        speed (float): character speed at the waypoint
        ease (Callable): easing function for character movement. Defaults to None.
    """

    waypoint_id: str
    coord: Coord
    speed: float = 1.0
    ease: typing.Callable | None = None
    layer: int = 0

    def __eq__(self, other: typing.Any) -> bool:
        if not isinstance(other, Waypoint):
            return NotImplemented
        return self.coord == other.coord

    def __hash__(self):
        return hash(self.waypoint_id)


class Motion:
    def __init__(self, character: "base_character.EffectCharacter"):
        """Motion class for managing the movement of a character.

        Args:
            character (base_character.EffectCharacter): The EffectCharacter to move.
        """
        self.waypoints: dict[str, Waypoint] = {}
        self.character = character
        self.current_coord: Coord = Coord(character.input_coord.column, character.input_coord.row)
        self.previous_coord: Coord = Coord(-1, -1)
        self.speed: float = 1.0
        self.active_waypoint: Waypoint | None = None
        self.origin_waypoint: Waypoint | None = None
        self.inter_waypoint_distance: float = 0
        self.inter_waypoint_max_steps: int = 0
        self.inter_waypoint_current_step: int = 0

    @staticmethod
    def find_coords_on_circle(origin: Coord, radius: int, num_points: int) -> list[Coord]:
        """Finds points on a circle.

        Args:
            origin (Coord): origin of the circle
            radius (int): radius of the circle
            num_points (int): number of points to find

        Returns:
            list (Coord): list of Coord points on the circle
        """
        points = []
        for i in range(num_points):
            angle = 2 * math.pi * i / num_points
            x = origin.column + radius * math.cos(angle)
            y = origin.row + radius * math.sin(angle)
            points.append(Coord(round(x), round(y)))
        return points

    @staticmethod
    def find_coords_in_circle(origin: Coord, radius: int, num_points: int) -> list[Coord]:
        """Finds points that fall within a circle with the given origin and radius. Points are
        chosen randomly from available points.

        Args:
            origin (Coord): origin of the circle
            radius (int): radius of the circle
            num_points (int): number of points to find

        Returns:
            list[Coord]: list of Coord points in the circle
        """
        points: list[Coord] = []
        selected_points: list[Coord] = []
        for i in range(1, radius + 1):
            points.extend(set((Motion.find_coords_on_circle(origin, i, 30))))
        while points and len(selected_points) < num_points:
            selected_points.append(points.pop(random.randrange(len(points))))
        return selected_points

    def _distance(self, column1: int, row1: int, column2: int, row2: int) -> float:
        """Returns the distance between two coordinates.

        Args:
            column1 (int): column of the first waypoint
            row1 (int): row of the first waypoint
            column2 (int): column of the second waypoint
            row2 (int): row of the second waypoint

        Returns:
            float: distance between the two waypoints
        """
        column_diff = column2 - column1
        row_diff = row2 - row1
        return math.hypot(column_diff, row_diff)

    def _point_at_distance(self, distance: float) -> Coord:
        """Returns the coordinate at the given distance along the line defined by the origin waypoint and current waypoint.

        Args:
            distance (float): distance

        Returns:
            Coord: Coordinate at the given distance.
        """
        if not distance or not self.origin_waypoint or not self.active_waypoint:
            return self.current_coord
        t = distance / self.inter_waypoint_distance
        next_column, next_row = (
            ((1 - t) * self.origin_waypoint.coord.column + t * self.active_waypoint.coord.column),
            ((1 - t) * self.origin_waypoint.coord.row + t * self.active_waypoint.coord.row),
        )
        return Coord(round(next_column), round(next_row))

    def set_coordinate(self, coord: Coord) -> None:
        """Sets the current coordinate to the given coordinate.

        Args:
            coord (Coord): coordinate
        """
        self.current_coord = coord

    def new_waypoint(
        self,
        waypoint_name: str,
        coord: Coord,
        *,
        speed: float = 1,
        ease: typing.Callable | None = None,
    ) -> Waypoint:
        """Creates a new Waypoint and appends adds it to the waypoints dictionary with the waypoint_name as key.

        Args:
            waypoint_name (str): unique identifier for the waypoint
            coord (Coord): coordinate
            speed (float): speed
            ease (Callable | None): easing function for character movement. Defaults to None.

        Returns:
            Waypoint: The new waypoint.
        """
        new_waypoint = Waypoint(waypoint_name, coord, speed, ease)
        self.waypoints[waypoint_name] = new_waypoint
        return new_waypoint

    def movement_is_complete(self) -> bool:
        """Returns whether the character has reached the final coordinate and moved the requisite number of steps.

        Returns:
            bool: True if the character has reached the final coordinate and has taken the maximum number of steps, False otherwise.
        """
        if self.active_waypoint is None:
            return True
        return False

    def _ease_movement(self, easing_func: typing.Callable) -> float:
        """Returns the percentage of total distance that should be moved based on the easing function.

        Args:
            easing_func (Callable): The easing function to use.

        Returns:
            float: The percentage of total distance to move.
        """

        elapsed_step_ratio = self.inter_waypoint_current_step / self.inter_waypoint_max_steps
        return easing_func(elapsed_step_ratio)

    def activate_waypoint(self, waypoint: Waypoint) -> None:
        """Sets the current waypoint to Waypoint and sets the speed, distance, and step values for the new waypoint. A
        WAYPOINT_ACTIVATED event is triggered.

        Args:
            waypoint (Waypoint): the Waypoint to activate
        """
        self.origin_waypoint = Waypoint("origin", Coord(self.current_coord.column, self.current_coord.row))
        self.active_waypoint = waypoint
        self.speed = self.active_waypoint.speed
        self.inter_waypoint_distance = self._distance(
            self.current_coord.column,
            self.current_coord.row,
            self.active_waypoint.coord.column,
            self.active_waypoint.coord.row,
        )
        self.inter_waypoint_current_step = 0
        if self.speed > self.inter_waypoint_distance:
            self.speed = max(self.inter_waypoint_distance, 1)
        self.inter_waypoint_max_steps = round(self.inter_waypoint_distance / self.speed)
        self.character.event_handler.handle_event(self.character.event_handler.Event.WAYPOINT_ACTIVATED, waypoint)

    def deactivate_waypoint(self, waypoint: Waypoint) -> None:
        """Unsets the current waypoint if the current waypoint is waypoint.

        Args:
            waypoint (Waypoint): the Waypoint to deactivate
        """
        if self.active_waypoint and self.active_waypoint is waypoint:
            self.active_waypoint = None
            self.inter_waypoint_distance = 0
            self.inter_waypoint_max_steps = 0
            self.inter_waypoint_current_step = 0

    def move(self) -> None:
        """Moves the character one step closer to the target position based on an easing function if present, otherwise linearly."""
        # preserve previous coordinate to allow for clearing the location in the terminal
        self.previous_coord = Coord(self.current_coord.column, self.current_coord.row)

        if not self.active_waypoint:
            return
        if self.inter_waypoint_current_step < self.inter_waypoint_max_steps:
            self.inter_waypoint_current_step += 1

        if self.inter_waypoint_distance:
            if self.active_waypoint and self.active_waypoint.ease:
                easing_factor = self._ease_movement(self.active_waypoint.ease)
                distance_to_move = easing_factor * self.inter_waypoint_distance
            else:
                linear_factor = self.inter_waypoint_current_step / self.inter_waypoint_max_steps
                distance_to_move = linear_factor * self.inter_waypoint_distance
            self.current_coord = self._point_at_distance(distance_to_move)

        if self.inter_waypoint_current_step == self.inter_waypoint_max_steps:
            waypoint_reached = self.active_waypoint
            self.deactivate_waypoint(self.active_waypoint)
            self.character.event_handler.handle_event(
                self.character.event_handler.Event.WAYPOINT_REACHED, waypoint_reached
            )
